"""
PC-BASIC - inputs package
Keyboard, pen and joystick handling

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .keyboard import Keyboard
from .pen import Pen
from .stick import Stick
