"""
PC-BASIC - devices package
Devices, Files and I/O operations

(c) 2013--2023 Rob Hagemans
This file is released under the GNU GPL version 3 or later.
"""

from .devicebase import TYPE_TO_MAGIC, InputTextFile
from .files import Files
from .disk import NameWrapper
