3 INPUT "screen"; S:SCREEN S
4 INPUT "char"; X$
5 CLS:LOCATE 1,1:PRINT X$;CHR$(219)
6     LOCATE 2,1:PRINT CHR$(219);CHR$(219)
7 LOCATE 1,10:PRINT "0         1     "
8 LOCATE 2,10:PRINT "0123456789012345"
10 FOR Y=0 TO 16
15 LOCATE 3+Y, 6:PRINT USING "##";Y
20 FOR X=0 TO 16
30 LOCATE 3+Y, 10+X
32 IF POINT(X,Y)=0 GOTO 40
35 PRINT"#"
40 NEXT
50 NEXT
