5 CLS
10 FOR BLINK = 0 TO 1
20 FOR BACK = 0 TO 7
30 FOR FG = 0 TO 15
40 ATTR = BLINK*128 + BACK*16 + FG
50 FGCOL = BLINK*16 + FG
60 COLOR FGCOL, BACK
70 PRINT USING "\\"; HEX$(ATTR);
80 NEXT
90 IF BACK MOD 2 = 1 THEN PRINT
100 NEXT: NEXT
110 COLOR 7,0
